/*
  Programa per a calcular les conques d'atracció del mètode iteratiu de Newton
per a una funció complexa.
  El programa escriu els punts de les conques d'atracció a la sortida estàndar;
per tant cal redirigir la sortida per a desar-ho a un arxiu i poder fer-ne
imatges amb el Gnuplot.

      ./conques-atraccio-complexos.out > arxiu_escriure_punts.dad
      

  Autor: Daniel Fernández Benítez.
  Última modificació: 18/06/2021.
  
  
  L'usuari proporciona una funció complexa, tots els seus zeros i un rectangle
en el pla complex.
  Aleshores es crea una xarxa de punts en el rectangle i cada punt es pren
com condició inicial del Mètode de Newton. Si els iterats de Newton s'atansen
a algun zero de la funció, aleshores la condició inicial pertany a la conca d'atracció de tal zero i escrivim aquesta condició inicial a la sortida estàndar,
seguint el següent format

    <part_real> <part_imaginària> <índex_zero>
    
on «índex_zero» és l'índex del zero en el vector de zeros que l'usuari
ha de donar.


-------------------------------
  Les dades que l'usuari pot tenir interès a modificar (funcio, constants, etc.)
es troben en XARXA_PUNTS i MÈTODE DE NEWTON.
-------------------------------


  El programa s'ha escrit definint com a macros del preprocessador aquells
càlculs que s'han de repetir molts cops i que, altrament, caldria escriure
en una funció. D'aquesta manera s'evita que el programa estigui constantment
fent crides a funcions, i així accelerem l'execució del programa.



POSSIBLES MILLORES
 - Potser fer els gràfics amb Gnuplot directament
  des del programa el mòdul  http://ndevilla.free.fr/gnuplot/
*/



#include<stdio.h>
#include<stdlib.h>
#include<complex.h>
#include<math.h>
#include<stdbool.h>

/*-----------------------------------------------------------
                      MÈTODE DE NEWTON
-----------------------------------------------------------*/

/*Informació relativa a l'execució del mètode de Newton:
condicions d'aturada, la funció i els seus zeros.*/


/*Nombre màxim d'iteracions*/
#define MAX_ITER 25

/*tolerància, al quadrat, en la proximitat a un zero de f(z) (TOL = 1E-3)*/
#define TOL2_CONCA 1E-6
/*tolerància, al quadrat, per considerar un nombre com 0 (TOL = 1E-12)*/
#define TOL2_NULLITAT 1E-24


/*funció a què aplicar el mètode de Newton*/
#define funcio(z) (z * z * z - 1.)
//#define funcio(z) (-3 + z * (7 + z * z * (-2 + z)))

/*derivada de la funció a què aplicar el mètode de Newton*/
#define funcio_1(z) (3 * z * z)
//#define funcio_1(z) (7 + z * z * (-6 + 4 * z))

/*quantitat de zeros que té funcio(z)*/
#define Q_ZEROS 3
//#define Q_ZEROS 4


/*Zeros de la funció definida dalt*/
//arrel quadrada de 3
#define AQ_3 1.73205080756887729352
double complex const g_zeros[Q_ZEROS] = {
    1.,
    -1. / 2. + AQ_3 / 2. * I,
    -1. / 2. - AQ_3 / 2. * I
};

/*Zeros de la funcio definida dalt*/
/*
double complex const g_zeros[Q_ZEROS] = {
  1.564432070603465 + 1.338737397039180 * I,
  1.564432070603465 - 1.338737397039180 * I,
  -1.577440734126577,
  .4485765929196463
};
*/
/*-----------------------------------------------------------
                   FI MÈTODE DE NEWTON
-----------------------------------------------------------*/


/*-----------------------------------------------------------
                        XARXA PUNTS
-----------------------------------------------------------*/

/*Domini rectangular en què crear la xarxa de punts*/
#define X_MIN -2.
#define Y_MAX 2.
#define X_MAX 2.
#define Y_MIN -2.

/*quantitat de punts en què dividir cada costat del rectangle*/
#define Q_PUNTS_XARXA_X 512
#define Q_PUNTS_XARXA_Y Q_PUNTS_XARXA_X

/*distància (horitzontal i vertical) entre punts de la xarxa*/
#define XARXA_PAS_X  ((X_MAX - X_MIN) / (Q_PUNTS_XARXA_X - 1))
#define XARXA_PAS_Y  ((Y_MAX - Y_MIN) / (Q_PUNTS_XARXA_Y - 1))
/*-----------------------------------------------------------
                      FI XARXA PUNTS
-----------------------------------------------------------*/


/*mòdul al quadrat d'un complex (cabs(z) ^ 2)*/
#define cmodul2(z) (creal(z) * creal(z) + cimag(z) * cimag(z))



char * g_nom_prog;


void conques_atraccio();




int main(int argc, char *argv[]){
  unsigned int i;
    
    
  /*processar arguments del programa*/
  g_nom_prog = argv[0];


  /*calcular i escriure les conques d'atracció*/
  conques_atraccio();
  
  
  /*escrivim també els zeros de la funció*/
  for(i = 0; i < Q_ZEROS; i++){
    printf("%13.7le %13.7le %u\n", creal(g_zeros[i]), cimag(g_zeros[i]), Q_ZEROS);
  }
   
    
  return EXIT_SUCCESS;
}


/*
DESC: Calcula a quina conca d'atracció pertany cada punt de la xarxa,
i ho escriu a la sortida estàndar.
  Amb cada punt de la xarxa, s'engega el mètode Newton i es comprova si s'atansa
a algun dels zeros de la funció. En cas afirmatiu escrivim a la sortida estàndar
el punt i l'índex del zero (índex en el vector «g_zeros»).

PARAMS:

RETORN:
*/
void conques_atraccio(){
  unsigned int i, j;
  double complex z, z_0;
  unsigned int conca;
  const double dx = XARXA_PAS_X, dy = XARXA_PAS_Y;
  
  
  /*Engeguem el mètode de Newton per a cada punt de la xarxa
  (d'esquerra a dreta i de dalt a baix)
  i mirem cap a quin zero s'atansa*/
  for(i = 0; i < Q_PUNTS_XARXA_Y; i++){
    for(j = 0; j < Q_PUNTS_XARXA_X; j++){
      unsigned int q_iter;
      bool es_conca_trobada;
      double complex f_1;
      
      
      /*mètode de Newton amb condició inicial*/
      z_0 = z = (X_MIN + j * dx) + (Y_MAX - i * dy) * I;
      es_conca_trobada = false;
      q_iter = 0;
      while(!es_conca_trobada && q_iter < MAX_ITER){
        /*Saltem el punt si la derivada en «z» s'anul·la*/
        f_1 = funcio_1(z);
        if(cmodul2(f_1) < TOL2_NULLITAT){
          break;
        }
        
        /*iteració Newton i comprovar si és prop d'un zero*/
        z = z - funcio(z) / f_1;
        
        for(conca = 0; !es_conca_trobada && conca < Q_ZEROS; conca++){
          if(cmodul2(g_zeros[conca] - z) < TOL2_CONCA){
            es_conca_trobada = true;
          }
        }
        if(es_conca_trobada){
          conca--;
        }
        
        
        q_iter++;
      }
      
      
      /*escrivim la dada inicial i la conca a què pertany*/
      if(es_conca_trobada){
        printf("%13.7le %13.7le %u\n", creal(z_0), cimag(z_0), conca);
      }
    }
  }
}

